@@ -39,7 +39,7 @@ pub fn expand_derive_request(input: DeriveInput) -> syn::Result<TokenStream> {
                 RequestField::Header(..) => collect_lifetime_idents(&mut lifetimes.header, ty),
                 RequestField::Body(_) => collect_lifetime_idents(&mut lifetimes.body, ty),
                 RequestField::NewtypeBody(_) => collect_lifetime_idents(&mut lifetimes.body, ty),
-                RequestField::NewtypeRawBody(_) => collect_lifetime_idents(&mut lifetimes.body, ty),
+                RequestField::RawBody(_) => collect_lifetime_idents(&mut lifetimes.body, ty),
                 RequestField::Path(_) => collect_lifetime_idents(&mut lifetimes.path, ty),
                 RequestField::Query(_) => collect_lifetime_idents(&mut lifetimes.query, ty),
                 RequestField::QueryMap(_) => collect_lifetime_idents(&mut lifetimes.query, ty),
@@ -119,12 +119,14 @@ impl Request {
         self.fields.iter().filter_map(RequestField::as_body_field)
     }
 
-    fn query_fields(&self) -> impl Iterator<Item = &Field> {
-        self.fields.iter().filter_map(RequestField::as_query_field)
+    fn has_body_fields(&self) -> bool {
+        self.fields
+            .iter()
+            .any(|f| matches!(f, RequestField::Body(_) | RequestField::NewtypeBody(_)))
     }
 
-    fn has_body_fields(&self) -> bool {
-        self.fields.iter().any(|f| matches!(f, RequestField::Body(..)))
+    fn has_newtype_body(&self) -> bool {
+        self.fields.iter().any(|f| matches!(f, RequestField::NewtypeBody(_)))
     }
 
     fn has_header_fields(&self) -> bool {
@@ -132,11 +134,11 @@ impl Request {
     }
 
     fn has_path_fields(&self) -> bool {
-        self.fields.iter().any(|f| matches!(f, RequestField::Path(..)))
+        self.fields.iter().any(|f| matches!(f, RequestField::Path(_)))
     }
 
     fn has_query_fields(&self) -> bool {
-        self.fields.iter().any(|f| matches!(f, RequestField::Query(..)))
+        self.fields.iter().any(|f| matches!(f, RequestField::Query(_)))
     }
 
     fn has_lifetimes(&self) -> bool {
@@ -154,12 +156,8 @@ impl Request {
         self.fields.iter().filter(|f| matches!(f, RequestField::Path(..))).count()
     }
 
-    fn newtype_body_field(&self) -> Option<&Field> {
-        self.fields.iter().find_map(RequestField::as_newtype_body_field)
-    }
-
-    fn newtype_raw_body_field(&self) -> Option<&Field> {
-        self.fields.iter().find_map(RequestField::as_newtype_raw_body_field)
+    fn raw_body_field(&self) -> Option<&Field> {
+        self.fields.iter().find_map(RequestField::as_raw_body_field)
     }
 
     fn query_map_field(&self) -> Option<&Field> {
@@ -172,17 +170,10 @@ impl Request {
         let ruma_serde = quote! { #ruma_api::exports::ruma_serde };
         let serde = quote! { #ruma_api::exports::serde };
 
-        let request_body_def = if let Some(body_field) = self.newtype_body_field() {
-            let field = Field { ident: None, colon_token: None, ..body_field.clone() };
-            Some(quote! { (#field); })
-        } else if self.has_body_fields() {
+        let request_body_struct = self.has_body_fields().then(|| {
+            let serde_attr = self.has_newtype_body().then(|| quote! { #[serde(transparent)] });
             let fields = self.fields.iter().filter_map(RequestField::as_body_field);
-            Some(quote! { { #(#fields),* } })
-        } else {
-            None
-        };
 
-        let request_body_struct = request_body_def.map(|def| {
             // Though we don't track the difference between newtype body and body
             // for lifetimes, the outer check and the macro failing if it encounters
             // an illegal combination of field attributes, is enough to guarantee
@@ -199,7 +190,8 @@ impl Request {
                     #serde::Serialize,
                     #derive_deserialize
                 )]
-                struct RequestBody< #(#lifetimes),* > #def
+                #serde_attr
+                struct RequestBody< #(#lifetimes),* > { #(#fields),* }
             }
         });
 
@@ -246,7 +238,7 @@ impl Request {
         // TODO: highlight problematic fields
 
         let newtype_body_fields = self.fields.iter().filter(|field| {
-            matches!(field, RequestField::NewtypeBody(_) | RequestField::NewtypeRawBody(_))
+            matches!(field, RequestField::NewtypeBody(_) | RequestField::RawBody(_))
         });
 
         let has_newtype_body_field = match newtype_body_fields.count() {
@@ -273,8 +265,8 @@ impl Request {
             }
         };
 
-        let has_body_fields = self.body_fields().count() > 0;
-        let has_query_fields = self.query_fields().count() > 0;
+        let has_body_fields = self.fields.iter().any(|f| matches!(f, RequestField::Body(_)));
+        let has_query_fields = self.fields.iter().any(|f| matches!(f, RequestField::Query(_)));
 
         if has_newtype_body_field && has_body_fields {
             return Err(syn::Error::new_spanned(
@@ -321,7 +313,7 @@ enum RequestField {
     NewtypeBody(Field),
 
     /// Arbitrary bytes in the body of the request.
-    NewtypeRawBody(Field),
+    RawBody(Field),
 
     /// Data that appears in the URL path.
     Path(Field),
@@ -342,7 +334,7 @@ impl RequestField {
                 RequestField::Header(field, header.expect("missing header name"))
             }
             RequestFieldKind::NewtypeBody => RequestField::NewtypeBody(field),
-            RequestFieldKind::NewtypeRawBody => RequestField::NewtypeRawBody(field),
+            RequestFieldKind::RawBody => RequestField::RawBody(field),
             RequestFieldKind::Path => RequestField::Path(field),
             RequestFieldKind::Query => RequestField::Query(field),
             RequestFieldKind::QueryMap => RequestField::QueryMap(field),
@@ -351,27 +343,34 @@ impl RequestField {
 
     /// Return the contained field if this request field is a body kind.
     pub fn as_body_field(&self) -> Option<&Field> {
-        self.field_of_kind(RequestFieldKind::Body)
-    }
-
-    /// Return the contained field if this request field is a body kind.
-    pub fn as_newtype_body_field(&self) -> Option<&Field> {
-        self.field_of_kind(RequestFieldKind::NewtypeBody)
+        match self {
+            RequestField::Body(field) | RequestField::NewtypeBody(field) => Some(field),
+            _ => None,
+        }
     }
 
     /// Return the contained field if this request field is a raw body kind.
-    pub fn as_newtype_raw_body_field(&self) -> Option<&Field> {
-        self.field_of_kind(RequestFieldKind::NewtypeRawBody)
+    pub fn as_raw_body_field(&self) -> Option<&Field> {
+        match self {
+            RequestField::RawBody(field) => Some(field),
+            _ => None,
+        }
     }
 
     /// Return the contained field if this request field is a query kind.
     pub fn as_query_field(&self) -> Option<&Field> {
-        self.field_of_kind(RequestFieldKind::Query)
+        match self {
+            RequestField::Query(field) => Some(field),
+            _ => None,
+        }
     }
 
     /// Return the contained field if this request field is a query map kind.
     pub fn as_query_map_field(&self) -> Option<&Field> {
-        self.field_of_kind(RequestFieldKind::QueryMap)
+        match self {
+            RequestField::QueryMap(field) => Some(field),
+            _ => None,
+        }
     }
 
     /// Gets the inner `Field` value.
@@ -380,26 +379,12 @@ impl RequestField {
             RequestField::Body(field)
             | RequestField::Header(field, _)
             | RequestField::NewtypeBody(field)
-            | RequestField::NewtypeRawBody(field)
+            | RequestField::RawBody(field)
             | RequestField::Path(field)
             | RequestField::Query(field)
             | RequestField::QueryMap(field) => field,
         }
     }
-
-    /// Gets the inner `Field` value if it's of the provided kind.
-    fn field_of_kind(&self, kind: RequestFieldKind) -> Option<&Field> {
-        match (self, kind) {
-            (RequestField::Body(field), RequestFieldKind::Body)
-            | (RequestField::Header(field, _), RequestFieldKind::Header)
-            | (RequestField::NewtypeBody(field), RequestFieldKind::NewtypeBody)
-            | (RequestField::NewtypeRawBody(field), RequestFieldKind::NewtypeRawBody)
-            | (RequestField::Path(field), RequestFieldKind::Path)
-            | (RequestField::Query(field), RequestFieldKind::Query)
-            | (RequestField::QueryMap(field), RequestFieldKind::QueryMap) => Some(field),
-            _ => None,
-        }
-    }
 }
 
 impl TryFrom<Field> for RequestField {
@@ -428,7 +413,7 @@ impl TryFrom<Field> for RequestField {
             field_kind = Some(match meta {
                 Meta::Word(ident) => match &ident.to_string()[..] {
                     "body" => RequestFieldKind::NewtypeBody,
-                    "raw_body" => RequestFieldKind::NewtypeRawBody,
+                    "raw_body" => RequestFieldKind::RawBody,
                     "path" => RequestFieldKind::Path,
                     "query" => RequestFieldKind::Query,
                     "query_map" => RequestFieldKind::QueryMap,
@@ -476,7 +461,7 @@ enum RequestFieldKind {
     Body,
     Header,
     NewtypeBody,
-    NewtypeRawBody,
+    RawBody,
     Path,
     Query,
     QueryMap,
